"""This package defines the `picharsso` command.

Refer to https://kelvindecosta.github.io/picharsso/commands/.
"""


import click

from .draw import draw
from .info import info
from ..meta import DESCRIPTION

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(
    help=DESCRIPTION,
    options_metavar="[options]",
    subcommand_metavar="<command> [args]",
    context_settings=CONTEXT_SETTINGS,
)
def main():
    """The main program."""


main.add_command(draw)
main.add_command(info)



"""This module defines the `picharsso info` command.

Refer to https://kelvindecosta.github.io/picharsso/commands/info/.
"""

from pathlib import Path

import click

from .. import __file__ as root_module_path
from ..meta import NAME, VERSION, DESCRIPTION, REPO_URL, DOCS_URL, LICENSE, AUTHOR
from ..utils import embolden, italicize


@click.command(options_metavar="[options]")
def info():
    """Displays package information."""
    with open(Path(root_module_path).parent / "data" / "logo.txt", "r") as file_stream:
        logo_text = file_stream.read().strip()

    output = logo_text.split("\n")

    line = 3
    output[line] += f"    {embolden(NAME)}"

    line += 2
    output[line] += f"    {italicize(DESCRIPTION)}"

    line += 3
    output[line] += f"    {italicize('Version')} : {VERSION}"

    line += 1
    output[line] += f"    {italicize('License')} : {LICENSE}"

    line += 1
    output[line] += f"    {italicize('Author')}  : {AUTHOR}"

    line += 3
    output[line] += f"    {italicize('Source')}  : {REPO_URL}"

    line += 1
    output[line] += f"    {italicize('Docs')}    : {DOCS_URL}"

    print("\n".join(output[1:-1]))



"""This module defines utility functions that are used across the package."""

from os import system, name
import shutil

import numpy as np
from PIL import Image
from sty import ef, rs


def clear_screen():
    """Clears the terminal console."""
    # Windows systems
    if name == "nt":
        _ = system("cls")
    # Unix systems
    else:
        _ = system("clear")


def embolden(text):
    """Modifies text to appear in a bold typeface,
    using [ANSI escape codes](https://en.wikipedia.org/wiki/ANSI_escape_code).

    Args:
        text (str): The subject text.

    Returns:
        str: The text in a bold typeface.
    """
    return f"{ef.bold}{text}{rs.bold_dim}"


def ensure_rgb(image):
    """Usually converts any [Pillow](https://python-pillow.org/)
    `image` to its equivalent in the `RGB` mode.

    Args:
        image (PIL.Image.Image): The subject image.

    Returns:
        PIL.Image.Image: The image in the `RGB` mode.
    """
    # If the image has a color palette,
    # convert to the `RGBA` mode.
    if image.mode == "P":
        image = image.convert("RGBA")

    # If the image is in `RGBA` mode,
    # create a white background.
    if image.mode == "RGBA":
        temp = Image.new("RGB", image.size, (255, 255, 255))
        temp.paste(image, mask=image.split()[3])
        image = temp

    # Convert to `RGB` mode
    return image.convert("RGB")


def italicize(text):
    """Modifies text to appear in italics,
    using [ANSI escape codes](https://en.wikipedia.org/wiki/ANSI_escape_code).

    Args:
        text (str): The subject text.

    Returns:
        str: The text in italics.
    """
    return f"{ef.italic}{text}{rs.italic}"


def submatrices(matrix, shape):
    """Returns a rolling window view of a `matrix`, without overlapping,
    given the `shape` of the window.

    Args:
        matrix (numpy.ndarray): The subject matrix.
        shape (Tuple[int, int]): The `<height>` and `<width>` of the window.

    Returns:
        numpy.ndarray: The rolling window view of the matrix.

    Note:
        This operation doesn't account for the loss of border elements.
    """
    # Extract strides and shapes for calculation.
    mat_hs, mat_ws = matrix.strides[:2]
    mat_h, mat_w = matrix.shape[:2]
    ker_h, ker_w = shape

    # View `matrix` according to new strides and shape.
    return np.lib.stride_tricks.as_strided(
        matrix,
        (1 + (mat_h - ker_h) // ker_h, 1 + (mat_w - ker_w) // ker_w, ker_h, ker_w)
        + matrix.shape[2:],
        strides=(ker_h * mat_hs, ker_w * mat_ws, mat_hs, mat_ws) + matrix.strides[2:],
    )


def terminal_size():
    """Returns the size of the terminal window.

    Returns:
        (Tuple[int, int]): The `<height>` and `<width>`
                            of the terminal window in characters.

    Note:
        When used while piping,
        this function usually returns the default terminal size, `(24, 80)`.
    """
    return shutil.get_terminal_size()[::-1]


__all__ = [
    "clear_screen",
    "embolden",
    "ensure_rgb",
    "italicize",
    "submatrices",
    "terminal_size",
]



"""This package defines the `picharsso draw` command.

Refer to https://kelvindecosta.github.io/picharsso/commands/draw/.
"""

import click
from PIL import Image

from ...draw import RESAMPLING_FILTERS, DEFAULT_RESAMPLING
from ...format import FORMATTERS, DEFAULT_FORMATTER
from ...utils import terminal_size

from .gradient import draw_gradient
from .braille import draw_braille


@click.group(options_metavar="[options]", subcommand_metavar="<command> [args]")
@click.argument("path", type=click.Path(exists=True), metavar="<path>")
@click.option(
    "-c", "--colorize", is_flag=True, help="Apply image colors to output text."
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(list(FORMATTERS.keys())),
    default=DEFAULT_FORMATTER,
    help="Format mode for output text.",
    show_default=True,
)
@click.option(
    "-r",
    "--resample",
    type=click.Choice(list(RESAMPLING_FILTERS.keys())),
    default=DEFAULT_RESAMPLING,
    help="Resampling filter.",
    show_default=True,
)
@click.option(
    "-H",
    "--height",
    type=int,
    default=0,
    help="Height of output text in characters.\n\nIf 0, derives from width.",
    show_default=True,
)
@click.option(
    "-W",
    "--width",
    type=int,
    default=0,
    help="Width of output text in characters.\n\nIf 0, derives from height.",
    show_default=True,
)
@click.option(
    "-term-h",
    "--terminal-height",
    is_flag=True,
    help="Sets height to terminal height.",
)
@click.option(
    "-term-w", "--terminal-width", is_flag=True, help="Sets width to terminal width.",
)
@click.pass_context
def draw(
    context,
    path,
    colorize,
    mode,
    resample,
    height,
    width,
    terminal_height,
    terminal_width,
):
    """Generate text art from an image.

        <path>                          Path to the image file.
    """
    image = Image.open(path)

    if terminal_width or terminal_height or height == 0 and width == 0:
        term_h, term_w = terminal_size()

        if terminal_height:
            height = term_h

        if terminal_width:
            width = term_w

        if height == 0 and width == 0:
            height = term_h
            width = term_w

    context.obj = {
        "image": image,
        "colorize": colorize,
        "mode": mode,
        "resample": resample,
        "height": height,
        "width": width,
    }


draw.add_command(draw_gradient)
draw.add_command(draw_braille)



"""This module defines variables for the package metadata."""

NAME: str = "picharsso"
"""The package name."""

DESCRIPTION: str = "A utility for converting images to text art."
"""The package description."""

VERSION: str = "2.0.1"
"""The current package version.
"""

REPO_URL: str = f"https://github.com/kelvindecosta/{NAME}"
"""The URL of the package source code."""

DOCS_URL: str = f"https://kelvindecosta.github.io/{NAME}"
"""The URL of the package documentation."""

AUTHOR: str = "Kelvin DeCosta"
"""The package author."""

LICENSE: str = "MIT"
"""The package license."""



"""This module defines a formatter for the [ANSI coloring scheme](../../formats/ansi.md)."""

from sty import fg

from .base import BaseFormatter


class AnsiFormatter(BaseFormatter):
    """
    A formatter for the [ANSI coloring scheme](../../formats/ansi.md).

    Inherits [`BaseFormatter`][picharsso.format.base.BaseFormatter].
    """

    @staticmethod
    def color(text, color):
        return f"{fg(*color)}{text}{fg.rs}"

    @staticmethod
    def translate(text_matrix):
        return text_matrix

    @staticmethod
    def unify(text_matrix):
        return "\n".join(["".join(row) for row in text_matrix])


__all__ = ["AnsiFormatter"]



"""This module defines the behaviour of the package
in a top-level script environment.

The following command executes this script:

```bash
python -m picharsso
```
"""

from .cli import main

if __name__ == "__main__":
    main()



"""
This module defines a drawer for the [Braille style](../../styles/braille.md).

??? example "Example"
    Consider the following image:

    <div align="center">
        <p>
            <img alt="Apple logo" src="../../assets/images/subjects/apple.webp" />
        </p>
        <p>
            <em>Apple Computer [Rob Janoff, 1977]</em>
        </p>
    </div>

    Here's what it should look like:

    <div align="center">
        <img
            alt="Apple logo in text (Braille style)"
            src="../../assets/images/outputs/demo/apple-braille.webp"
        />
    </div>
"""

import numpy as np

from .base import BaseDrawer
from ..utils import submatrices

DEFAULT_THRESHOLD: int = 64
"""The default threshold grayscale intensity."""


class BrailleDrawer(BaseDrawer):
    """
    A drawer for the [Braille style](../../styles/braille.md).

    Inherits [`BaseDrawer`][picharsso.draw.base.BaseDrawer].

    Attributes:
        threshold (int): Threshold grayscale intensity for pixels to be considered.
        kernel (numpy.ndarray): A hard-coded matrix relating the intensity to
                             the Unicode values for Braille characters.
        charset_array (numpy.ndarray): A matrix of all Braille characters,
                                    indexed by their offsetted Unicode value.
    """

    def __init__(self, threshold=DEFAULT_THRESHOLD, **kwargs):
        """Initialization method.

        Args:
            threshold (Optional[int]): Threshold grayscale intensity
                                    for pixels to be considered.
        """

        super().__init__(**kwargs)
        self.threshold = None
        self.set(threshold=threshold)

        self.kernel = np.array([[1, 8], [2, 16], [4, 32], [64, 128]]).astype(np.uint8)
        self.charset_array = np.array([chr(ord("\u2800") + x) for x in range(256)])

    def calculate_size(self, image_size):
        # Possible dimensions
        new_h = self.height
        new_w = self.width

        new_h = new_h * 4
        new_w = new_w * 2

        # Image dimensions
        old_h, old_w = image_size

        # If height is not set, infer it from width
        if not new_h:
            new_h = int(round(old_h / old_w * new_w))

        # If width is not set, infer it from height
        if not new_w:
            new_w = int(round(old_w / old_h * new_h))

        return new_h, new_w

    def process(self, image):
        # Convert the image mode to grayscale.
        # Filter all pixels with intensity greater than or equal to the threshold.
        # Perform a convolution on this filtered image with the Braille kernel.
        # The resultant matrix has the offsetted Unicode values, i.e., indices
        # for the corresponding Braille characters that form the image.
        # Index the character set with the indices.
        return self.charset_array[
            np.einsum(
                "ij,klij->kl",
                self.kernel,
                submatrices(
                    (np.array(image.convert("L")) >= self.threshold).astype(np.uint8),
                    self.kernel.shape,
                ),
            )
        ]

    def set(self, threshold=None, **kwargs):
        """Sets attributes of the drawer instance.

        Args:
            threshold (Optional[int]): Sets `threshold`.
            **kwargs (dict): Appropriate keyword arguments.
                    See [`BaseDrawer.set`][picharsso.draw.base.BaseDrawer.set].
        """
        super().set(**kwargs)

        if threshold is not None:
            self.threshold = threshold


__all__ = ["BrailleDrawer"]



"""This module defines the `picharsso draw braille` command.

Refer to https://kelvindecosta.github.io/picharsso/commands/draw/braille/.
"""


import click

from ...draw import new_drawer
from ...draw.braille import DEFAULT_THRESHOLD


@click.command("braille", options_metavar="[options]")
@click.option(
    "-t",
    "--threshold",
    type=click.IntRange(0, 255),
    help="Threshold pixel luminance (from grayscale).",
    default=DEFAULT_THRESHOLD,
    show_default=True,
)
@click.pass_context
def draw_braille(context, threshold):
    """Use the Braille style."""

    image = context.obj.pop("image")

    drawer = new_drawer("braille", threshold=threshold, **context.obj)
    print(drawer(image))



"""This module defines the `picharsso draw gradient` command.

Refer to https://kelvindecosta.github.io/picharsso/commands/draw/gradient/.
"""

import click

from ...draw import new_drawer
from ...draw.gradient import DEFAULT_CHARSET


@click.command("gradient", options_metavar="[options]")
@click.option(
    "-s",
    "--charset",
    type=str,
    help="Character set ordered by increasing 'brightness'.",
    default=DEFAULT_CHARSET,
    show_default=True,
)
@click.option(
    "-n", "--negative", is_flag=True, help="Whether to invert output text brightness."
)
@click.pass_context
def draw_gradient(context, charset, negative):
    """Use the gradient style."""

    image = context.obj.pop("image")

    drawer = new_drawer("gradient", charset=charset, negative=negative, **context.obj)
    print(drawer(image))



"""
This module defines a drawer for the [gradient style](../../styles/gradient.md).

??? example "Example"
    Consider the following image:

    <div align="center">
        <p>
            <img alt="Apple logo" src="../../assets/images/subjects/apple.webp" />
        </p>
        <p>
            <em>Apple Computer [Rob Janoff, 1977]</em>
        </p>
    </div>

    Here's what it should look like:

    <div align="center">
        <img
            alt="Apple logo in text (gradient style)"
            src="../../assets/images/outputs/demo/apple-gradient.webp"
        />
    </div>
"""

import numpy as np

from .base import BaseDrawer

DEFAULT_CHARSET: str = " :!?PG@"
"""The default character set."""


class GradientDrawer(BaseDrawer):
    """
    A drawer for the [gradient style](../../styles/gradient.md).

    Inherits [`BaseDrawer`][picharsso.draw.base.BaseDrawer].

    Attributes:
        charset (str): A set of characters ordered by the amount of area
                        their symbols occupy.
        negative (bool): Whether or not to reverse the `charset`.
        charset_array (numpy.ndarray): A vectorized version of the `charset`.
    """

    def __init__(self, charset=DEFAULT_CHARSET, negative=False, **kwargs):
        """Initialization method.

        Args:
            charset (Optional[str]): A set of characters ordered
                                    by the amount of area their symbols occupy.
                                    Defaults to `DEFAULT_CHARSET`
            negative (Optional[bool]): Whether or not to reverse the `charset`.
            **kwargs (dict): Appropriate keyword arguments.
                            See [`BaseDrawer`][picharsso.draw.base.BaseDrawer].
        """
        super().__init__(**kwargs)
        self.charset = None
        self.negative = None
        self.charset_array = None
        self.set(charset=charset, negative=negative)

    def calculate_size(self, image_size):
        # Possible dimensions
        new_h = self.height
        new_w = self.width

        # Image dimensions
        old_h, old_w = image_size

        # If height is not set, infer it from width
        if not new_h:
            new_h = int(round(old_h / old_w * new_w / 2.125))

        # If width is not set, infer it from height
        if not new_w:
            new_w = int(round(old_w / old_h * new_h * 2.125))

        return new_h, new_w

    def process(self, image):
        # Convert the image mode to grayscale.
        # Normalize the pixel values from a range of (0, 255) to (0, len(self.charset)-1),
        # to obtain indices for the character set.
        # Index the character set array with the indices.
        return self.charset_array[
            np.round(
                np.array(image.convert("L")) / 255 * (len(self.charset) - 1)
            ).astype(int)
        ]

    def set(self, charset=None, negative=None, **kwargs):
        """Sets attributes of the drawer instance.

        Args:
            charset (Optional[str]): Sets `charset`.
            negative (Optional[bool]): Sets `negative`.
            **kwargs (dict): Appropriate keyword arguments.
                    See [`BaseDrawer.set`][picharsso.draw.base.BaseDrawer.set].
        """
        super().set(**kwargs)

        if charset is not None:
            self.charset = charset

        if negative is not None:
            self.negative = negative

        self.charset_array = np.array(
            list(self.charset if not self.negative else self.charset[::-1])
        )


__all__ = ["GradientDrawer"]



"""
This module defines an abstract base formatter.

!!! question "Formats"
    Refer to the [Formats documentation](../../formats/index.md)
    to learn about the supported output formats.
"""

from abc import ABC, abstractmethod

import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured


class BaseFormatter(ABC):
    """
    An abstract base formatter.

    Attributes:
        colorize (bool): Whether to color the text.
        vcolor (Callable): The vectorized implementation of the `color` method.

    Note:
        The following methods must be overwritten:

        - [`color`][picharsso.format.base.BaseFormatter.color]
        - [`translate`][picharsso.format.base.BaseFormatter.translate]
        - [`unify`][picharsso.format.base.BaseFormatter.unify]
    """

    def __init__(self, colorize=False):
        """Initialization method.

        Args:
            colorize (Option[bool]): Whether to color the text.
        """
        self.colorize = None
        BaseFormatter.set(self, colorize=colorize)

        self.vcolor = np.vectorize(self.color)

    def __call__(self, text_matrix, image, resample):
        """Applies formatting and colorization on the `text_matrix`
        and returns a single string.

        Args:
            text_matrix (numpy.ndarray): The subject text matrix,
                                        with `shape = (<height>, <width>)`,
                                        and `dtype = str`.
            image (PIL.Image.Image): The subject image.
            resample (int): The resampling filter.

        Returns:
            str: The formatted string of text with color (if specified).
        """

        text_size = text_matrix.shape

        # Apply any translations.
        text_matrix = self.translate(text_matrix)

        # Colorize if necessary
        if self.colorize:
            # Pool the colors from the original image by resizing it to the size of the text output.
            # Using the vectorized `color` method, color each element in the `text_martix`.
            # The vectorized operation takes a `str` from `text_matrix`
            # and a `List[int, int, int]` from the pooled colors.
            text_matrix = self.vcolor(
                text_matrix,
                unstructured_to_structured(
                    np.array(image.resize(text_size[::-1], resample=resample)).astype(
                        np.uint8
                    )
                ).astype("O"),
            )

        return self.unify(text_matrix)

    @staticmethod
    @abstractmethod
    def color(text, color):
        """Applies `color` to a string of `text`.

        Args:
            text (str): The subject text.
            color (Tuple[int, int, int]): The `RGB` value for the color.

        Returns:
            str: The colored text.
        """

    @staticmethod
    @abstractmethod
    def translate(text_matrix):
        """Applies translatations to `text_matrix`.

        Args:
            text_matrix (numpy.ndarray): The subject text matrix,
                                        with `shape = (<height>, <width>)`,
                                        and `dtype = str`.

        Returns:
            numpy.ndarray: The translated text_matrix.
        """

    @staticmethod
    @abstractmethod
    def unify(text_matrix):
        """Formats a `text_matrix` into a single string.

        Args:
            text_matrix (numpy.ndarray): The subject text matrix,
                                        with `shape = (<height>, <width>)`,
                                        and `dtype = str`.

        Returns:
            str: The formatted string of text art.
        """

    def set(self, colorize=None):
        """Sets attributes of the formatter instance.

        Args:
            colorize (Optional[bool]): Sets `colorize`.
        """
        if colorize is not None:
            self.colorize = colorize


__all__ = ["BaseFormatter"]



"""
This module defines an abstract base drawer.

!!! question "Styles"
    Refer to the [Styles documentation](../../styles/index.md)
    for an in-depth guide to the **image processing behind Picharsso**.
"""

from abc import ABC, abstractmethod

from PIL import Image

from ..format import new_formatter
from ..utils import ensure_rgb

RESAMPLING_FILTERS: dict = {
    "nearest": Image.NEAREST,
    "box": Image.BOX,
    "bilinear": Image.BILINEAR,
    "hamming": Image.HAMMING,
    "bicubic": Image.BICUBIC,
    "lanczos": Image.LANCZOS,
}
"""A collection of resampling filters.
See [Pillow's Filters](https://pillow.readthedocs.io/en/stable/handbook/concepts.html#filters){target=_blank}.
"""

DEFAULT_RESAMPLING: str = "nearest"
"""The default resampling filter."""


class BaseDrawer(ABC):
    """
    An abstract base drawer.

    Attributes:
        height (int): The desired height of the text in characters.
        width (int): The desired width of the text in characters.
        resample (int): The resampling filter.
        format (Type[picharsso.format.BaseFormatter]): The formatter instance.

    Note:
        The following methods must be overwritten:

        - [`calculate_size`][picharsso.draw.base.BaseDrawer.calculate_size]
        - [`process`][picharsso.draw.base.BaseDrawer.process]
    """

    def __init__(self, height=42, width=0, resample=DEFAULT_RESAMPLING, **kwargs):
        """Initialization method.

        Args:
            height (Optional[int]): The desired height of the text in characters.
            width (Optional[int]): The desired width of the text in characters.
            resample (Optional[str]): The resampling filter.
            **kwargs (dict): Appropriate keyword arguments.
                    See [`BaseFormatter`][picharsso.format.base.BaseFormatter] and others.

        Note:
            When set as `0`, `height` is derived from `width` and vice versa.
            This is done to preserve the aspect ratio of the image.
        """
        self.height = None
        self.width = None
        self.resample = None
        BaseDrawer.set(self, height=height, width=width, resample=resample)

        self.format = new_formatter(**kwargs)

    def __call__(self, image):
        """Applies processing and formatting on the `image`
        and returns a single string.

        Args:
            image (PIL.Image.Image): The subject image.

        Returns:
            str: The string of text art.
        """
        # Ensure that the image is in the `RGB` mode.
        image = ensure_rgb(image)

        # Calculate the new size of the image, for processing the text matrix.
        image_size = self.calculate_size(image.size[::-1])

        # Process text matrix from the resized image.
        text_matrix = self.process(
            image.resize(image_size[::-1], resample=self.resample)
        )

        # Apply formatting.
        return self.format(text_matrix, image, self.resample)

    @abstractmethod
    def calculate_size(self, image_size):
        """Calculates the size of the image for processing the text matrix.

        Args:
            image_size (Tuple[int, int]): The height and width of the subject image.

        Returns:
            Tuple[int, int]: The size of the image.
        """

    @abstractmethod
    def process(self, image):
        """Converts an image to a matrix of text.

        Args:
            image (PIL.Image.Image): The subject image,
                                    with `mode = "RGB"`,
                                    and `size = (<height>, <width>)`.

        Returns:
            numpy.ndarray: The text matrix,
                        with `shape = (<height>, <width>)`,
                        and `dtype = str`.
        """

    def set(self, height=None, width=None, resample=None):
        """Sets attributes of the drawer instance.

        Args:
            height (Option[int]): Sets `height`.
            width (Option[int]): Sets `width`.
            resample (Option[str]): Sets `resample`.

        Raises:
            ValueError: If both `height` and `width` are set to `0`.
        """
        # Set resampling filter
        if resample is not None:
            self.resample = RESAMPLING_FILTERS[resample]

        # Set height and width
        if height is not None or width is not None:
            new_h = self.height if height is None else height
            new_w = self.width if width is None else width

            if new_h == 0 and new_w == 0:
                raise ValueError("Either height or width must be non-zero")

            self.height = new_h
            self.width = new_w


__all__ = ["BaseDrawer", "RESAMPLING_FILTERS", "DEFAULT_RESAMPLING"]



from setuptools import find_packages, setup

from src.meta import NAME, DESCRIPTION, VERSION, REPO_URL, DOCS_URL, AUTHOR, LICENSE

with open("README.md", "r") as fs:
    LONGE_DESCRIPTION = fs.read()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONGE_DESCRIPTION,
    long_description_content_type="text/markdown",
    project_urls={"Documentation": DOCS_URL, "Source": REPO_URL},
    author=AUTHOR,
    author_email="decostakelvin@gmail.com",
    license=LICENSE,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Artistic Software",
        "Topic :: Terminals",
        "Topic :: Utilities",
    ],
    packages=list(map(lambda x: x.replace("src", NAME), find_packages("."))),
    package_dir={NAME: "src"},
    package_data={"": ["data/*.txt"]},
    python_requires=">=3.8",
    install_requires=["click", "numpy", "pillow", "sty",],
    entry_points={"console_scripts": [f"{NAME} = {NAME}.cli:main"]},
)



"""
This package defines drawers for different styles of text art.

!!! question "Styles"
    Refer to the [Styles documentation](../../styles/index.md)
    for an in-depth guide to the **image processing behind Picharsso**.
"""

from .gradient import GradientDrawer
from .base import BaseDrawer, RESAMPLING_FILTERS, DEFAULT_RESAMPLING
from .braille import BrailleDrawer

DRAWERS: dict = {"gradient": GradientDrawer, "braille": BrailleDrawer}
"""The collection of drawers."""


def new_drawer(style, **kwargs):
    """Creates a new drawer instance.

    Args:
        style (str): The style of the text art.
        **kwargs (dict): Appropriate keyword arguments.
                        See [`BaseDrawer`][picharsso.draw.base.BaseDrawer]
                        and others.

    Returns:
        Type[picharsso.draw.BaseDrawer]: The new drawer instance.
    """
    return DRAWERS[style](**kwargs)



"""
This package defines formatters for different modes of text output.

!!! question "Formats"
    Refer to the [Formats documentation](../../formats/index.md)
    to learn about the supported output formats.
"""

from .ansi import AnsiFormatter
from .base import BaseFormatter
from .html import HtmlFormatter

FORMATTERS: dict = {"ansi": AnsiFormatter, "html": HtmlFormatter}
"""The collection of formatters."""

DEFAULT_FORMATTER: str = "ansi"
"""The default formatter."""


def new_formatter(mode=DEFAULT_FORMATTER, **kwargs):
    """Creates a new formatter instance.

    Args:
        mode (Option[str]): The mode of the output text.
                            Defaults to `DEFAULT_FORMATTER`.
        **kwargs (dict): Appropriate keyword arguments.
                        See [`BaseFormatter`][picharsso.format.base.BaseFormatter]
                        and others.

    Returns:
        Type[picharsso.format.BaseFormatter] : The new formatter instance.
    """
    return FORMATTERS[mode](**kwargs)



"""This module defines a formatter for [HTML](../../formats/html.md)."""

from html.entities import name2codepoint

import numpy as np

from .base import BaseFormatter

HTML_ENTITY_MAP: dict = {chr(value): key for (key, value) in name2codepoint.items()}
"""A dictionary mapping unicode characters to their equivalent HTML entities."""

HTML_ENTITY_MAP[" "] = "nbsp;"


class HtmlFormatter(BaseFormatter):
    """
    A formatter for [HTML](../../formats/html.md).

    Inherits [`BaseFormatter`][picharsso.format.base.BaseFormatter].
    """

    @staticmethod
    def color(text, color):
        return f'<span style="color : rgb{tuple(color)};">{text}</span>'

    @staticmethod
    def translate(text_matrix):
        unique_chars = np.unique(text_matrix)

        # Change datatype to accomodate strings of varying length
        text_matrix = text_matrix.astype(
            f"<U{len(max(name2codepoint.keys(), key=len))}"
        )

        # Change each character to its equivalent entity
        for char in unique_chars:
            if char in HTML_ENTITY_MAP:
                text_matrix[text_matrix == char] = f"&{HTML_ENTITY_MAP[char]}"

        return text_matrix

    @staticmethod
    def unify(text_matrix):
        return "<div>{}</div>".format(
            "\n".join([f"<div>{''.join(row)}</div>" for row in text_matrix])
        )


__all__ = ["HtmlFormatter", "HTML_ENTITY_MAP"]
